# encoding = utf-8

import os
import sys
sys.path.insert(0, '..')

import numpy as np
from PIL import Image
from torch.utils.data import Dataset
from torch.utils.data import DataLoader
import torch
from tqdm import tqdm
from collections import OrderedDict


class LRUCache:
    def __init__(self, capacity: int):
        self.cache = OrderedDict()
        self.capacity = capacity

    def get(self, key):
        if key not in self.cache:
            return None
        else:
            self.cache.move_to_end(key)
            return self.cache[key]

    def put(self, key, value) -> None:
        self.cache[key] = value
        self.cache.move_to_end(key)
        if len(self.cache) > self.capacity:
            self.cache.popitem(last=False)


TRAIN = "train"
TEST = "test"

MEAN = 255 * torch.tensor([0.485, 0.456, 0.406])
STD = 255 * torch.tensor([0.229, 0.224, 0.225])
PROJECT_DIR = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
DATA_BASE_DIR = os.path.join(os.path.dirname(PROJECT_DIR), 'movingDetection/data')


class CalipsaDataset(Dataset):

    # ground truth labels: {"1": 1, "2": 2, "person": 3, "vehicle": 4}
    # CLASS2ID = {"1": 0, "2": 1, "person": 2, "vehicle": 3}

    def __init__(self, split, data_base=DATA_BASE_DIR, transforms=None, min_size=35):
        assert(split in {"train", "test"})
        self.split = split
        self.data_base_dir = data_base
        self.transforms = transforms

        # augmentation flags
        self.flip_flag = False

        # Load proposal boxes
        self.proposals = load_proposal_boxes(data_base, split, len(self), min_size=min_size)

        # Load Targets and Labels
        self.targets = self.load_all_targets()

        # Load Test Ground Truth Boxes
        if self.split == TEST:
            self.test_gt_boxes = self.load_all_test_gt_boxes()
            self.test_gt_labels = self.load_all_test_gt_labels()
            assert(len(self.test_gt_boxes) == len(self.test_gt_labels))

        # Images Cache (size: 500 pairs)
        self.cache = LRUCache(5)

    def enable_augmentation(self):
        pass

    def load_all_test_gt_labels(self):
        gt_filepath = os.path.join(self.data_base_dir, "test_gt_labels.txt")
        gt = []
        with open(gt_filepath, mode='r', encoding='utf-8') as f:
            lines = f.readlines()
            assert(len(lines) == len(self))
            for line in lines:
                if line == '\n':
                    gt.append(None)
                else:
                    gt.append(torch.IntTensor([int(e) - 1 for e in line.rstrip().split(' ')]))
        return gt

    def load_all_targets(self):
        """
        :return: Tensor[length x num_classes]
        """
        targets_dir = os.path.join(self.data_base_dir, f"{self.split}_targets")
        batch_targets_list = []
        for idx in range(100, len(self) + 1, 100):
            batch_targets_list.append(torch.load(os.path.join(targets_dir, f"{self.split}_targets_{idx}.pt")))
        return torch.cat(batch_targets_list, dim=0)

    def load_all_test_gt_boxes(self):
        gt_filepath = os.path.join(self.data_base_dir, "test_gt_boxes.txt")
        gt = []
        with open(gt_filepath, mode='r', encoding='utf-8') as f:
            lines = f.readlines()
            single_gt = []
            first_flg, gt_num = True, 0
            for line in lines:
                if line == "GTBOXES\n":
                    gt_num += 1
                    if not first_flg:
                        if not single_gt:
                            gt.append(None)
                        else:
                            gt.append(torch.FloatTensor(single_gt))
                        single_gt = []
                    else:
                        first_flg = False
                else:
                    _tmp = [float(e) for e in line.rstrip().split(' ')]
                    _tmp[0], _tmp[2] = 640 * _tmp[0], 640 * _tmp[2]
                    _tmp[1], _tmp[3] = 480 * _tmp[1], 480 * _tmp[3]
                    single_gt.append(_tmp)

            assert(gt_num == len(self))
            assert(len(gt) == len(self) - 1)
            if single_gt:
                gt.append(torch.FloatTensor(single_gt))

            else:
                gt.append(None)
            assert(len(gt) == len(self))
        return gt

    def __len__(self):
        if self.split == TRAIN:
            return 27000
        elif self.split == TEST:
            return 3000

    def __getitem__(self, index):
        item, img_scale, hflip, batch_idx = index
        batch_imgs_name = f"{self.split}_imgs_{(int(item / 100 + 1) * 100)}.pt"
        batch_imgs = self.cache.get(batch_imgs_name)
        if batch_imgs is None:
            batch_imgs = torch.load(os.path.join(self.data_base_dir, f"{self.split}_images", batch_imgs_name))
            self.cache.put(batch_imgs_name, batch_imgs)
        specific_imgs = batch_imgs[item % 100].float()
        to_transform = {'img': specific_imgs.permute(1, 2, 0).numpy(), 'proposals': self.proposals[item].numpy(),
                        'img_scale': img_scale, 'hflip': hflip}
        if self.transforms is not None:
            to_transform = self.transforms(to_transform)
        img = to_transform['img']
        proposals = to_transform['proposals']

        if self.split == TRAIN:
            return img, self.targets[item].float(), proposals, item, self.proposals[item]

        elif self.split == TEST:
            return img, self.targets[item].float(), proposals, self.proposals[item], \
                   self.test_gt_labels[item], self.test_gt_boxes[item]
        else:
            raise Exception("Only TRAIN and TEST split!")


class CalipsaDataLoader(object):
    def __init__(self, dataset, batch_size=1):
        self.batch_size = batch_size
        self.dataset = dataset
        self.length = len(self.dataset)

    def __len__(self):
        if self.length % self.batch_size == 0:
            return int(self.length / self.batch_size)
        else:
            return int(self.length/self.batch_size) + 1

    def __iter__(self):
        self.idx = 0
        return self

    def __next__(self):
        if self.idx < self.length:
            tuple_list = []
            for _ in range(min(self.batch_size, self.length - self.idx)):
                tuple_list.append(self.dataset.__getitem__(self.idx))
                self.idx += 1

            res_list = []
            for j in range(3):
                res_list.append(torch.stack([t[j] for t in tuple_list]))
            res_list.append([t[3] for t in tuple_list])

            if self.dataset.split == TEST:
                res_list.append([t[-2] for t in tuple_list])
                res_list.append([t[-1] for t in tuple_list])

            return res_list
        else:
            raise StopIteration


def load_proposal_boxes(proposal_base, split, length, min_size=30):
    """
    load proposal boxes into main memory
    :param min_size:      filter small boxes
    :param proposal_base: str
    :param split:         "train" or "test"
    :param length:        length of img pairs
    :return:              List[Tensor[L x 4]]
    """

    proposal_dir = f"{split}_proposals_edgebox"
    total_boxes = []
    for file_idx in range(100, length + 1, 100):
        filename = f"{split}_proposal_boxes_{file_idx}.npy"
        batch_boxes_np = np.load(os.path.join(proposal_base, proposal_dir, filename))
        if proposal_dir == f"{split}_proposals_edgebox":
            batch_boxes_np[:, :, 2] = batch_boxes_np[:, :, 0] + batch_boxes_np[:, :, 2]
            batch_boxes_np[:, :, 3] = batch_boxes_np[:, :, 1] + batch_boxes_np[:, :, 3]
        if split == TRAIN:
            for boxes_idx in range(100):
                boxes_np = batch_boxes_np[boxes_idx, :, :]
                nonzero_boxes_mask = filter_small_boxes(boxes_np, min_size=3)
                boxes_filtered_np = boxes_np[nonzero_boxes_mask]
                total_boxes.append(torch.from_numpy(boxes_filtered_np).float())
        else:
            for boxes_idx in range(100):
                boxes_np = batch_boxes_np[boxes_idx, :, :]
                large_boxes_mask = filter_small_boxes(boxes_np, min_size=min_size)
                boxes_filtered_np = boxes_np[large_boxes_mask]
                total_boxes.append(torch.from_numpy(boxes_filtered_np).float())
    assert(len(total_boxes) == length)
    return total_boxes


def filter_small_boxes(boxes, min_size=20):
    """Filters out small boxes."""
    w = boxes[:, 2] - boxes[:, 0]
    h = boxes[:, 3] - boxes[:, 1]
    mask = (w >= min_size) & (h >= min_size)
    return mask


def normalized_(img):
    assert(img.shape[0] == 3)
    img = (img - MEAN[:, None, None]) / STD[:, None, None]
    return img


if __name__ == '__main__':
    train_ds = CalipsaDataset("test")
    train_dl = CalipsaDataLoader(train_ds, batch_size=2)
    cnt = 0
    for batch_imgs, batch_target, batch_boxes in train_dl:
        if cnt > 0: break
        cnt += 1
        print(batch_boxes.shape)